import string


def is_printable(ch: str):
    return (
        ch in string.digits or ch in string.ascii_letters or ch in string.punctuation or ch == " "
    )


def print_byte(b: int) -> str:
    ch = chr(b)
    return ch if is_printable(ch) else "."


def split_data(data: bytearray, size: int):
    for i in range(0, len(data), size):
        yield data[i : i + size]


def dump_data_str(data: bytearray) -> str:
    return "".join([print_byte(b) for b in data])


def dump_data(data: bytearray, input: bool, debug: bool) -> bytearray:
    if debug:
        dir = "<-" if input else "->"
        text = "\n   ".join(
            [
                f"{chunk.hex(' ').ljust(48)}# {dump_data_str(chunk)}"
                for chunk in split_data(data, 16)
            ]
        )
        print(f"{dir} {text}")
    return data


def dump_in(data: bytearray, debug: bool) -> bytearray:
    return dump_data(data, True, debug)


def dump_out(data: bytearray, debug: bool) -> bytearray:
    return dump_data(data, False, debug)



from __future__ import annotations

from bleak import BleakScanner
from bleak.backends.device import BLEDevice

RADONEYE_NAME_PREFIX = "FR:"


class RadonEyeScanner:
    @classmethod
    async def discover(
        cls,
        timeout: float = 5,
        adapter: str | None = None,
    ) -> list[BLEDevice]:
        devices = await BleakScanner.discover(timeout=timeout, adapter=adapter)  # type: ignore
        return [dev for dev in devices if dev.name and dev.name.startswith(RADONEYE_NAME_PREFIX)]



from abc import abstractmethod
from typing import Literal, TypedDict

RadonUnit = Literal["bq/m3", "pci/l"]

OutputType = Literal["text", "json"]


class RadonEyeStatus(TypedDict):
    serial: str
    model: str
    firmware_version: str
    latest_bq_m3: float
    latest_pci_l: float
    day_avg_bq_m3: float
    day_avg_pci_l: float
    month_avg_bq_m3: float
    month_avg_pci_l: float
    peak_bq_m3: float
    peak_pci_l: float
    counts_current: int
    counts_previous: int
    counts_str: str
    uptime_minutes: int
    uptime_str: str
    display_unit: RadonUnit
    alarm_enabled: int
    alarm_level_bq_m3: float
    alarm_level_pci_l: float
    alarm_interval_minutes: int


class RadonEyeHistory(TypedDict):
    values_bq_m3: list[float]
    values_pci_l: list[float]


class RadonEyeInterface:
    @abstractmethod
    def supports(self) -> bool:
        raise NotImplementedError("Not supported method supports()")

    @abstractmethod
    async def status(self) -> RadonEyeStatus:
        raise NotImplementedError("Not supported method status()")

    @abstractmethod
    async def history(self) -> RadonEyeHistory:
        raise NotImplementedError("Not supported method history()")

    @abstractmethod
    async def beep(self) -> None:
        raise NotImplementedError("Not supported method beep()")

    @abstractmethod
    async def set_alarm(
        self,
        enabled: bool,
        level: float,
        unit: RadonUnit,
        interval: int,
    ) -> None:
        raise NotImplementedError("Not supported method set_alarm()")

    @abstractmethod
    async def set_unit(self, unit: RadonUnit) -> None:
        raise NotImplementedError("Not supported method set_unit()")



from __future__ import annotations

import asyncio
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Literal, NamedTuple, TypedDict

from radoneye.client import RadonEyeClient
from radoneye.model import OutputType, RadonUnit
from radoneye.scanner import RadonEyeScanner
from radoneye.util import convert_radon_value, serialize_object


class RadonEyeAlarmStatus(TypedDict):
    alarm_enabled: int
    alarm_level_bq_m3: float
    alarm_level_pci_l: float
    alarm_interval_minutes: int


class ListCommandArgs(NamedTuple):
    adapter: str | None
    timeout: int
    output: OutputType


async def cmd_list(args: ListCommandArgs):
    devs = [
        {"address": dev.address, "name": dev.name}
        for dev in await RadonEyeScanner.discover(adapter=args.adapter, timeout=args.timeout)
    ]
    if args.output == "text":
        for dev in devs:
            print(f"{dev['address']}\t{dev['name']}")
    else:
        print(serialize_object(devs, "json"))


class BeepCommandArgs(NamedTuple):
    adapter: str | None
    debug: bool
    connect_timeout: int
    address: str


async def cmd_beep(args: BeepCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
        debug=args.debug,
    ) as client:
        print("beep")
        await client.beep()


class StatusCommandArgs(NamedTuple):
    adapter: str | None
    debug: bool
    connect_timeout: int
    read_timeout: int
    output: OutputType
    address: str


async def cmd_status(args: StatusCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
        status_read_timeout=args.read_timeout,
        debug=args.debug,
    ) as client:
        status = await client.status()
        print(serialize_object(status, args.output))


class HistoryCommandArgs(NamedTuple):
    adapter: str | None
    debug: bool
    connect_timeout: int
    read_timeout: int
    output: OutputType
    address: str


async def cmd_history(args: HistoryCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
        history_read_timeout=args.read_timeout,
        debug=args.debug,
    ) as client:
        history = await client.history()
        if args.output == "text":
            print("#\tBq/m3\tpCi/L")
            for index, value_bq_m3 in enumerate(history["values_bq_m3"]):
                value_pci_l = history["values_pci_l"][index]
                print(f"{index + 1}\t{value_bq_m3}\t{value_pci_l}")
        else:
            print(serialize_object(history, "json"))


class AlarmCommandArgs(NamedTuple):
    adapter: str | None
    debug: bool
    connect_timeout: int
    read_timeout: int
    output: OutputType
    address: str
    status: Literal["on", "off"] | None
    level: float | None  # in bq/m3 or pci/l
    unit: RadonUnit | None
    interval: int | None  # mins


async def cmd_alarm(args: AlarmCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
        status_read_timeout=args.read_timeout,
        debug=args.debug,
    ) as client:
        if args.status is None and args.level is None and args.interval is None:
            status = await client.status()

            alarm_status = {
                "alarm_enabled": status["alarm_enabled"],
                "alarm_level_bq_m3": status["alarm_level_bq_m3"],
                "alarm_level_pci_l": status["alarm_level_pci_l"],
                "alarm_interval_minutes": status["alarm_interval_minutes"],
            }

            print(serialize_object(alarm_status, args.output))
        elif (
            args.status is not None
            and args.level is not None
            and args.unit is not None
            and args.interval is not None
        ):
            new_alarm_status: RadonEyeAlarmStatus = {
                "alarm_enabled": args.status == "on",
                "alarm_level_bq_m3": convert_radon_value(args.level, args.unit, "bq/m3"),
                "alarm_level_pci_l": convert_radon_value(args.level, args.unit, "pci/l"),
                "alarm_interval_minutes": args.interval,
            }

            await client.set_alarm(
                enabled=args.status == "on",
                level=args.level,
                unit=args.unit,
                interval=args.interval,
            )

            print(serialize_object(new_alarm_status, args.output))
        else:
            status = await client.status()

            new_enabled = (
                args.status == "on" if args.status is not None else bool(status["alarm_enabled"])
            )

            if args.level is not None:
                new_unit = args.unit or status["display_unit"]
                new_level = args.level
            elif status["display_unit"] == "pci/l":
                new_level = status["alarm_level_pci_l"]
                new_unit = "pci/l"
            else:
                new_level = status["alarm_level_bq_m3"]
                new_unit = "bq/m3"

            new_interval = args.interval or status["alarm_interval_minutes"]

            await client.set_alarm(
                enabled=new_enabled,
                level=new_level,
                unit=new_unit,
                interval=new_interval,
            )

            new_alarm_status: RadonEyeAlarmStatus = {
                "alarm_enabled": new_enabled,
                "alarm_level_bq_m3": convert_radon_value(new_level, new_unit, "bq/m3"),
                "alarm_level_pci_l": convert_radon_value(new_level, new_unit, "pci/l"),
                "alarm_interval_minutes": new_interval,
            }

            print(serialize_object(new_alarm_status, args.output))


class UnitCommandArgs(NamedTuple):
    adapter: str | None
    debug: bool
    connect_timeout: int
    read_timeout: int
    output: OutputType
    address: str
    unit: RadonUnit | None


async def cmd_unit(args: UnitCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
        status_read_timeout=args.read_timeout,
        debug=args.debug,
    ) as client:
        if args.unit is None:
            status = await client.status()
            print(serialize_object(status["display_unit"], args.output))
        else:
            await client.set_unit(args.unit)
            print(serialize_object(args.unit, args.output))


async def main(argv: list[str]):
    parser = ArgumentParser(
        description="Ecosense RadonEye command line interface (currently supports RD200 v1/v2)",
        formatter_class=ArgumentDefaultsHelpFormatter,
        prog="radoneye",
    )

    parser.add_argument("--adapter", type=int, help="bluetooth adapter name (hci0 on Linux)")
    parser.add_argument(
        "-d", "--debug", action="store_true", default=False, help="enable to see message dumps"
    )

    subparsers = parser.add_subparsers(required=True, help="sub-command help")

    parser_list = subparsers.add_parser(
        "list",
        help="scan for devices available nearby",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_list.add_argument("--timeout", type=int, help="scan timeout", default=5)
    parser_list.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_list.set_defaults(func=cmd_list)

    parser_beep = subparsers.add_parser(
        "beep",
        help="trigger beep",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_beep.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_beep.add_argument("address", help="device address")
    parser_beep.set_defaults(func=cmd_beep)

    parser_status = subparsers.add_parser(
        "status",
        help="read status",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_status.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_status.add_argument("--read-timeout", type=int, help="read timeout", default=5)
    parser_status.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_status.add_argument("address", help="device address")
    parser_status.set_defaults(func=cmd_status)

    parser_history = subparsers.add_parser(
        "history",
        help="read history",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_history.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_history.add_argument("--read-timeout", type=int, help="read timeout", default=60)
    parser_history.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_history.add_argument("address", help="device address")
    parser_history.set_defaults(func=cmd_history)

    parser_alarm = subparsers.add_parser(
        "alarm",
        help="get/set alarm configuration",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_alarm.add_argument("address", help="device address")
    parser_alarm.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_alarm.add_argument("--read-timeout", type=int, help="read timeout", default=5)
    parser_alarm.add_argument("--status", choices=["on", "off"], help="alarm status")
    parser_alarm.add_argument("--level", type=float, help="alarm level in bq/m3 or pci/l")
    parser_alarm.add_argument("--unit", choices=["bq/m3", "pci/l"], help="alarm level unit")
    parser_alarm.add_argument("--interval", type=int, help="alarm interval (in minutes)")
    parser_alarm.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_alarm.set_defaults(func=cmd_alarm)

    parser_unit = subparsers.add_parser(
        "unit",
        help="get/set display unit (bq/m3 or pci/l)",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_unit.add_argument("address", help="device address")
    parser_unit.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_unit.add_argument("--read-timeout", type=int, help="read timeout", default=5)
    parser_unit.add_argument("--unit", choices=["bq/m3", "pci/l"], help="set new display unit")
    parser_unit.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_unit.set_defaults(func=cmd_unit)

    args = parser.parse_args(argv[1:])

    await args.func(args)


if __name__ == "__main__":
    asyncio.run(main(sys.argv))



from __future__ import annotations

import asyncio
import math

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic

from radoneye.debug import dump_in, dump_out
from radoneye.model import RadonEyeHistory, RadonEyeInterface, RadonEyeStatus, RadonUnit
from radoneye.util import (
    encode_bool,
    encode_byte,
    encode_float,
    format_counts,
    format_uptime,
    read_bool,
    read_byte,
    read_float,
    read_int,
    read_short,
    read_short_list,
    read_str_wl,
    round_pci_l,
    to_bq_m3,
    to_pci_l,
)

SERVICE_UUID = "00001523-1212-efde-1523-785feabcd123"

CHAR_COMMAND = "00001524-1212-efde-1523-785feabcd123"
CHAR_STATUS = "00001525-1212-efde-1523-785feabcd123"
CHAR_HISTORY = "00001526-1212-efde-1523-785feabcd123"

COMMAND_STATUS_10 = 0x10  # requests message A4, A8, AC, 50, 51
COMMAND_STATUS_A6 = 0xA6  # requests message A6
COMMAND_STATUS_AF = 0xAF  # requests message AF
COMMAND_STATUS_50 = 0x50  # requests message 50
COMMAND_STATUS_51 = 0x51  # requests message 51
COMMAND_STATUS_E8 = 0xE8  # requests message E8
COMMAND_HISTORY = 0xE9  # triggers history messages
COMMAND_BEEP = 0xA1  # triggers beep
COMMAND_SET_ALARM = 0xAA  # requests message AC, updates alarm settings
COMMAND_SET_UNIT = 0xA2  # requests message AC, updates unit settings

MSG_PREAMBLE_50 = 0x50  # latest/daily/monthly radon, raw particle counts
MSG_PREAMBLE_51 = 0x51  # uptime/peak radon
MSG_PREAMBLE_A4 = 0xA4  # manufacturing date/serial
MSG_PREAMBLE_A6 = 0xA6  # model series
MSG_PREAMBLE_A8 = 0xA8  # model name
MSG_PREAMBLE_AC = 0xAC  # alarm settings
MSG_PREAMBLE_AF = 0xAF  # software version
MSG_PREAMBLE_E8 = 0xE8  # history size

INVOKE_DELAY = 0.2  # sec


def parse_status(
    msg_50: bytearray,
    msg_51: bytearray,
    msg_a4: bytearray,
    msg_a6: bytearray,
    msg_a8: bytearray,
    msg_ac: bytearray,
    msg_af: bytearray,
) -> RadonEyeStatus:
    # Most messages start with command code followed by byte representing length of data inside buffer
    # buffer has at most 20 bytes (including command code), unused buffer part can contain "trash".

    serial_part1 = read_str_wl(msg_a6, 1)  # series?
    serial_part2 = read_str_wl(msg_a4, 1)[2:8]  # manufacturing date (YYMMDD)?
    serial_part3 = read_str_wl(msg_a4, 1)[-4:]  # serial within manufacturing date?
    serial = serial_part1 + serial_part2 + serial_part3  # example: {RU2}{201202}{0159}

    model = read_str_wl(msg_a8, 2)

    firmware_version = read_str_wl(msg_af, 1).rstrip()  # value has useless trailing new line

    latest_value = read_float(msg_50, 2)
    latest_pci_l = round_pci_l(latest_value)
    latest_bq_m3 = to_bq_m3(latest_value)

    day_avg_value = read_float(msg_50, 6)
    day_avg_pci_l = round_pci_l(day_avg_value)
    day_avg_bq_m3 = to_bq_m3(day_avg_value)

    month_avg_value = read_float(msg_50, 10)
    month_avg_pci_l = round_pci_l(month_avg_value)
    month_avg_bq_m3 = to_bq_m3(month_avg_value)

    counts_current = read_short(msg_50, 14)
    counts_previous = read_short(msg_50, 16)
    counts_str = format_counts(counts_current, counts_previous)

    uptime_minutes = read_int(msg_51, 4)
    uptime_str = format_uptime(uptime_minutes)

    peak_value = read_float(msg_51, 12)
    peak_pci_l = round_pci_l(peak_value)
    peak_bq_m3 = to_bq_m3(peak_value)

    display_unit = "bq/m3" if read_bool(msg_ac, 2) else "pci/l"

    alarm_enabled = read_bool(msg_ac, 3)
    alarm_level_value = read_float(msg_ac, 4)
    alarm_level_pci_l = round_pci_l(alarm_level_value)
    alarm_level_bq_m3 = to_bq_m3(alarm_level_pci_l)
    alarm_interval = read_byte(msg_ac, 8)
    alarm_interval_minutes = alarm_interval * 10

    return {
        "serial": serial,
        "model": model,
        "firmware_version": firmware_version,
        "latest_bq_m3": latest_bq_m3,
        "latest_pci_l": latest_pci_l,
        "day_avg_bq_m3": day_avg_bq_m3,
        "day_avg_pci_l": day_avg_pci_l,
        "month_avg_bq_m3": month_avg_bq_m3,
        "month_avg_pci_l": month_avg_pci_l,
        "peak_bq_m3": peak_bq_m3,
        "peak_pci_l": peak_pci_l,
        "counts_current": counts_current,
        "counts_previous": counts_previous,
        "counts_str": counts_str,
        "uptime_minutes": uptime_minutes,
        "uptime_str": uptime_str,
        "display_unit": display_unit,
        "alarm_enabled": alarm_enabled,
        "alarm_level_bq_m3": alarm_level_bq_m3,
        "alarm_level_pci_l": alarm_level_pci_l,
        "alarm_interval_minutes": alarm_interval_minutes,
    }


def parse_history_size(msg_e8: bytearray) -> int:
    return read_short(msg_e8, 2)


def parse_history_data(msg_e9: bytearray, size: int) -> RadonEyeHistory:
    values = read_short_list(msg_e9, 0, size)
    # 2.7 is experimentally found value that is used to scale down values back to pci/l
    values_pci_l = [v / 37 / 2.7 for v in values]
    return {
        "values_bq_m3": [to_bq_m3(v) for v in values_pci_l],
        "values_pci_l": [round_pci_l(v) for v in values_pci_l],
    }


class InterfaceV1(RadonEyeInterface):
    def __init__(
        self,
        client: BleakClient,
        status_read_timeout: float | None = None,
        history_read_timeout: float | None = None,
        debug: bool = False,
    ):
        self.client = client
        self.status_read_timeout = status_read_timeout
        self.history_read_timeout = history_read_timeout
        self.debug = debug

    def supports(self) -> bool:
        return bool(self.client.services.get_service(SERVICE_UUID))

    async def status(self) -> RadonEyeStatus:
        loop = asyncio.get_running_loop()

        future = loop.create_future()

        msg_50: bytearray | None = None
        msg_51: bytearray | None = None
        msg_a4: bytearray | None = None
        msg_a6: bytearray | None = None
        msg_a8: bytearray | None = None
        msg_ac: bytearray | None = None
        msg_af: bytearray | None = None

        def callback(char: BleakGATTCharacteristic, data: bytearray) -> None:
            nonlocal msg_50
            nonlocal msg_51
            nonlocal msg_a4
            nonlocal msg_a6
            nonlocal msg_a8
            nonlocal msg_ac
            nonlocal msg_af

            if data[0] == MSG_PREAMBLE_50:
                msg_50 = dump_in(data, self.debug)
            elif data[0] == MSG_PREAMBLE_51:
                msg_51 = dump_in(data, self.debug)
            elif data[0] == MSG_PREAMBLE_A4:
                msg_a4 = dump_in(data, self.debug)
            elif data[0] == MSG_PREAMBLE_A6:
                msg_a6 = dump_in(data, self.debug)
            elif data[0] == MSG_PREAMBLE_A8:
                msg_a8 = dump_in(data, self.debug)
            elif data[0] == MSG_PREAMBLE_AC:
                msg_ac = dump_in(data, self.debug)
            elif data[0] == MSG_PREAMBLE_AF:
                msg_af = dump_in(data, self.debug)

            if msg_50 and msg_51 and msg_a4 and msg_a6 and msg_a8 and msg_ac and msg_af:
                status = parse_status(
                    msg_50,
                    msg_51,
                    msg_a4,
                    msg_a6,
                    msg_a8,
                    msg_ac,
                    msg_af,
                )
                future.set_result(status)

        await self.client.start_notify(CHAR_STATUS, callback)  # type: ignore
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_STATUS_10]), self.debug)
        )
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_STATUS_AF]), self.debug)
        )
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_STATUS_A6]), self.debug)
        )
        result = await asyncio.wait_for(future, self.status_read_timeout)
        await self.client.stop_notify(CHAR_STATUS)
        return result

    async def history(self) -> RadonEyeHistory:
        loop = asyncio.get_running_loop()

        size_future = loop.create_future()
        result_future = loop.create_future()

        result_size: int = -1
        result_data = bytearray()

        def callback_status(char: BleakGATTCharacteristic, data: bytearray) -> None:
            nonlocal result_size
            if data[0] == MSG_PREAMBLE_E8:
                size_future.set_result(parse_history_size(dump_in(data, self.debug)))

        def callback_history(char: BleakGATTCharacteristic, data: bytearray) -> None:
            nonlocal result_data
            result_data.extend(dump_in(data, self.debug))
            if len(result_data) >= result_size * 2:
                result_future.set_result(parse_history_data(result_data, result_size))

        await self.client.start_notify(CHAR_STATUS, callback_status)  # type: ignore
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_STATUS_E8]), self.debug)
        )
        result_size = await asyncio.wait_for(size_future, self.status_read_timeout)
        await self.client.stop_notify(CHAR_STATUS)

        if result_size == 0:
            result = RadonEyeHistory(values_bq_m3=[], values_pci_l=[])
        else:
            await self.client.start_notify(CHAR_HISTORY, callback_history)  # type: ignore
            await self.client.write_gatt_char(
                CHAR_COMMAND, dump_out(bytearray([COMMAND_HISTORY]), self.debug)
            )
            result = await asyncio.wait_for(result_future, self.history_read_timeout)
            await self.client.stop_notify(CHAR_HISTORY)

        return result

    async def beep(self) -> None:
        # RadonEye app writes longer command, but it is actually enough to send one byte to beep
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_BEEP]), self.debug)
        )
        # there is some delay needed before you can do next beep, otherwise it will be just one beep
        await asyncio.sleep(INVOKE_DELAY)

    async def set_alarm(
        self,
        enabled: bool,
        level: float,
        unit: RadonUnit,
        interval: int,
    ) -> None:
        command = (
            bytearray([COMMAND_SET_ALARM, 0x11])
            + encode_bool(enabled)
            + encode_float(level if unit == "pci/l" else to_pci_l(level))
            + encode_byte(math.ceil(interval / 10))
        )
        await self.client.write_gatt_char(CHAR_COMMAND, dump_out(command, self.debug))
        await asyncio.sleep(INVOKE_DELAY)  # doesn't work without delay

    async def set_unit(self, unit: RadonUnit) -> None:
        command = bytearray([COMMAND_SET_UNIT, 0x11]) + encode_bool(
            False if unit == "pci/l" else True
        )
        await self.client.write_gatt_char(CHAR_COMMAND, dump_out(command, self.debug))
        await asyncio.sleep(INVOKE_DELAY)  # doesn't work without delay



from __future__ import annotations

import asyncio
import math
from typing import TypedDict

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic

from radoneye.debug import dump_in, dump_out
from radoneye.model import RadonEyeHistory, RadonEyeInterface, RadonEyeStatus, RadonUnit
from radoneye.util import (
    encode_bool,
    encode_byte,
    encode_short,
    format_counts,
    format_uptime,
    read_bool,
    read_byte,
    read_int,
    read_short,
    read_short_list,
    read_str,
    to_bq_m3,
    to_pci_l,
)

SERVICE_UUID = "00001523-0000-1000-8000-00805f9b34fb"

CHAR_COMMAND = "00001524-0000-1000-8000-00805f9b34fb"
CHAR_STATUS = "00001525-0000-1000-8000-00805f9b34fb"
CHAR_HISTORY = "00001526-0000-1000-8000-00805f9b34fb"

COMMAND_STATUS = 0x40
COMMAND_HISTORY = 0x41
COMMAND_BEEP = 0xA1
COMMAND_SET_ALARM = 0xAA
COMMAND_SET_UNIT = 0xA2

INVOKE_DELAY = 0.2  # sec


class RadonEyeHistoryPage(TypedDict):
    page_count: int
    page_no: int
    value_count: int
    values_bq_m3: list[int]
    values_pci_l: list[float]


def parse_status(data: bytearray) -> RadonEyeStatus:
    if read_byte(data, 15) == 0x06:  # v2
        serial_part1 = read_str(data, 8, 3)  # series?
        serial_part2 = read_str(data, 2, 6)  # manufacturing date (YYMMDD)?
        serial_part3 = read_str(data, 11, 4)  # serial within manufacturing date?
        serial = serial_part1 + serial_part2 + serial_part3
        model = read_str(data, 16, 6)
    elif read_byte(data, 14) == 0x07:  # v3
        serial = read_str(data, 2, 12)
        model = read_str(data, 15, 7)
    else:
        serial = "unknown"
        model = "unknown"

    firmware_version = read_str(data, 22, 6)

    display_unit = "bq/m3" if read_bool(data, 28) else "pci/l"

    alarm_enabled = read_bool(data, 29)
    alarm_level_bq_m3 = read_short(data, 30)
    alarm_level_pci_l = to_pci_l(alarm_level_bq_m3)
    alarm_interval = read_byte(data, 32)
    alarm_interval_minutes = alarm_interval * 10

    latest_bq_m3 = read_short(data, 33)
    latest_pci_l = to_pci_l(latest_bq_m3)

    day_avg_bq_m3 = read_short(data, 35)
    day_avg_pci_l = to_pci_l(day_avg_bq_m3)

    month_avg_bq_m3 = read_short(data, 37)
    month_avg_pci_l = to_pci_l(month_avg_bq_m3)

    counts_current = read_short(data, 39)
    counts_previous = read_short(data, 41)
    counts_str = format_counts(counts_current, counts_previous)

    uptime_minutes = read_int(data, 43)
    uptime_str = format_uptime(uptime_minutes)

    peak_bq_m3 = read_short(data, 51)
    peak_pci_l = to_pci_l(peak_bq_m3)

    return {
        "serial": serial,
        "model": model,
        "firmware_version": firmware_version,
        "latest_bq_m3": latest_bq_m3,
        "latest_pci_l": latest_pci_l,
        "day_avg_bq_m3": day_avg_bq_m3,
        "day_avg_pci_l": day_avg_pci_l,
        "month_avg_bq_m3": month_avg_bq_m3,
        "month_avg_pci_l": month_avg_pci_l,
        "peak_bq_m3": peak_bq_m3,
        "peak_pci_l": peak_pci_l,
        "counts_current": counts_current,
        "counts_previous": counts_previous,
        "counts_str": counts_str,
        "uptime_minutes": uptime_minutes,
        "uptime_str": uptime_str,
        "display_unit": display_unit,
        "alarm_enabled": alarm_enabled,
        "alarm_level_bq_m3": alarm_level_bq_m3,
        "alarm_level_pci_l": alarm_level_pci_l,
        "alarm_interval_minutes": alarm_interval_minutes,
    }


def parse_history_page(data: bytearray) -> RadonEyeHistoryPage:
    data.pop(0)  # command

    page_count = data.pop(0)
    page_no = data.pop(0)
    value_count = data.pop(0)
    values_bq_m3 = read_short_list(data, 0, len(data) // 2)
    values_pci_l = [to_pci_l(x) for x in values_bq_m3]

    return {
        "page_count": page_count,
        "page_no": page_no,
        "value_count": value_count,
        "values_bq_m3": values_bq_m3,
        "values_pci_l": values_pci_l,
    }


def merge_history(pages: list[RadonEyeHistoryPage]) -> RadonEyeHistory:
    pages = sorted(pages, key=lambda page: page["page_no"])
    if pages:
        for index, page in enumerate(pages):
            if index + 1 != page["page_no"]:
                raise ValueError("History page order mismatch")
        if pages[0]["page_count"] != len(pages):
            raise ValueError("History page count mismatch")
    return {
        "values_bq_m3": [value for page in pages for value in page["values_bq_m3"]],
        "values_pci_l": [value for page in pages for value in page["values_pci_l"]],
    }


class InterfaceV2(RadonEyeInterface):
    def __init__(
        self,
        client: BleakClient,
        status_read_timeout: float | None = None,
        history_read_timeout: float | None = None,
        debug: bool = False,
    ):
        self.client = client
        self.status_read_timeout = status_read_timeout
        self.history_read_timeout = history_read_timeout
        self.debug = debug

    def supports(self) -> bool:
        return bool(self.client.services.get_service(SERVICE_UUID))

    async def status(self) -> RadonEyeStatus:
        loop = asyncio.get_running_loop()
        future = loop.create_future()

        def callback(char: BleakGATTCharacteristic, data: bytearray) -> None:
            if data[0] == COMMAND_STATUS:
                future.set_result(parse_status(dump_in(data, self.debug)))

        await self.client.start_notify(CHAR_STATUS, callback)  # type: ignore
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_STATUS]), self.debug)
        )
        result = await asyncio.wait_for(future, self.status_read_timeout)
        await self.client.stop_notify(CHAR_STATUS)
        return result

    async def history(self) -> RadonEyeHistory:
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        pages: list[RadonEyeHistoryPage] = []

        def callback(char: BleakGATTCharacteristic, data: bytearray) -> None:
            if data[0] == COMMAND_HISTORY:
                page = parse_history_page(dump_in(data, self.debug))
                pages.append(page)
                if page["page_count"] == page["page_no"]:
                    future.set_result(merge_history(pages))

        await self.client.start_notify(CHAR_HISTORY, callback)  # type: ignore
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_HISTORY]), self.debug)
        )
        result = await asyncio.wait_for(future, self.history_read_timeout)
        await self.client.stop_notify(CHAR_HISTORY)
        return result

    async def beep(self) -> None:
        # RadonEye app writes longer command, but it is actually enough to send one byte to beep
        await self.client.write_gatt_char(
            CHAR_COMMAND, dump_out(bytearray([COMMAND_BEEP]), self.debug)
        )
        # there is some delay needed before you can do next beep, otherwise it will be just one beep
        await asyncio.sleep(INVOKE_DELAY)

    async def set_alarm(
        self,
        enabled: bool,
        level: float,
        unit: RadonUnit,
        interval: int,
    ) -> None:
        command = (
            bytearray([COMMAND_SET_ALARM, 0x11])
            + encode_bool(enabled)
            + encode_short(to_bq_m3(level) if unit == "pci/l" else level)
            + encode_byte(math.ceil(interval / 10))
        )
        await self.client.write_gatt_char(CHAR_COMMAND, dump_out(command, self.debug))
        await asyncio.sleep(INVOKE_DELAY)  # doesn't work without delay

    async def set_unit(self, unit: RadonUnit) -> None:
        command = bytearray([COMMAND_SET_UNIT, 0x11]) + encode_bool(
            False if unit == "pci/l" else True
        )
        await self.client.write_gatt_char(CHAR_COMMAND, dump_out(command, self.debug))
        await asyncio.sleep(INVOKE_DELAY)  # doesn't work without delay



from __future__ import annotations

import json
import math
import os
from struct import pack, unpack_from
from typing import Any, Literal, cast

from radoneye.model import RadonUnit

RADONEYE_ROUNDING_OFF = os.environ.get("RADONEYE_ROUNDING_OFF", "false") == "true"


def read_str_wl(buffer: bytearray, offset: int) -> str:
    # string length is encoded as first byte followed by string content with optional new line
    return buffer[(offset + 1) : (offset + 1 + buffer[offset])].decode()


def read_str(buffer: bytearray, offset: int, length: int) -> str:
    return buffer[(offset) : (offset + length)].decode()


def read_float(buffer: bytearray, offset: int) -> float:
    return float(unpack_from("<f", buffer, offset)[0])


def read_int(buffer: bytearray, offset: int) -> int:
    return unpack_from("<I", buffer, offset)[0]


def read_short(buffer: bytearray, offset: int) -> int:
    return unpack_from("<H", buffer, offset)[0]


def read_short_list(buffer: bytearray, offset: int, size: int) -> list[int]:
    return cast(list[int], unpack_from("<" + "H" * size, buffer, offset))


def encode_float(value: float) -> bytearray:
    return bytearray(pack("<f", value))


def encode_short(value: int | float) -> bytearray:
    value_to_pack = round(value) if isinstance(value, float) else value
    return bytearray(pack("<H", value_to_pack))


def read_bool(buffer: bytearray, offset: int) -> bool:
    return unpack_from("<c", buffer, offset)[0][0] == 0x01


def encode_bool(value: bool) -> bytearray:
    return bytearray([0x01 if value else 0x00])


def read_byte(buffer: bytearray, offset: int) -> int:
    return unpack_from("<c", buffer, offset)[0][0]


def encode_byte(value: int) -> bytearray:
    return bytearray(pack("<c", bytes([value])))


def round_pci_l(value_pci_l: float) -> float:
    if RADONEYE_ROUNDING_OFF:
        return value_pci_l
    return round(value_pci_l, 2)


def to_bq_m3(value_pci_l: float) -> float:
    if RADONEYE_ROUNDING_OFF:
        return float(value_pci_l * 37)
    return float(round(value_pci_l * 37))


def to_pci_l(value_bq_m3: float) -> float:
    return round_pci_l(value_bq_m3 / 37)


def convert_radon_value(value: float, from_unit: RadonUnit, to_unit: RadonUnit) -> float:
    if from_unit == to_unit:
        return value
    if from_unit == "bq/m3" and to_unit == "pci/l":
        return to_pci_l(value)
    if from_unit == "pci/l" and to_unit == "bq/m3":
        return to_bq_m3(value)
    raise ValueError(f"Unexpected conversion schema: {from_unit} -> {to_unit}")


def format_uptime(uptime_minutes: int) -> str:
    uptime_days = math.floor(uptime_minutes / (60 * 24))
    uptime_hours = math.floor(uptime_minutes % (60 * 24) / 60)
    uptime_mins = uptime_minutes % 60
    return f"{uptime_days}d{uptime_hours:02}h{uptime_mins:02}m"


def format_counts(counts_current: int, counts_previous: int) -> str:
    return f"{counts_current}/{counts_previous}"


def serialize_value_text(v: Any):
    if isinstance(v, bool):
        return "yes" if v else "no"
    return v


def serialize_object(obj: Any, output: Literal["text", "json"]):
    if output == "text":
        if isinstance(obj, dict):  # type: ignore
            return "\n".join(
                [
                    f"{key}\t{serialize_value_text(obj.get(key))}"  # type: ignore
                    for key in sorted(obj.keys())  # type: ignore
                ]
            )
        else:
            return f"{serialize_value_text(obj)}"
    else:
        return json.dumps(obj, separators=(",", ":"))



from __future__ import annotations

from typing import Union

from bleak import BleakClient
from bleak.backends.device import BLEDevice

from radoneye.interface_v1 import InterfaceV1
from radoneye.interface_v2 import InterfaceV2
from radoneye.model import RadonEyeHistory, RadonEyeInterface, RadonEyeStatus, RadonUnit


class RadonEyeClient:
    def __init__(
        self,
        address_or_ble_device: Union[BLEDevice, str],
        connect_timeout: float = 10,
        status_read_timeout: float = 5,
        history_read_timeout: float = 60,
        adapter: str | None = None,
        debug: bool = False,
    ) -> None:
        self.client = BleakClient(address_or_ble_device, timout=connect_timeout, adapter=adapter)
        self.interface: RadonEyeInterface | None = None
        self.status_read_timeout = status_read_timeout
        self.history_read_timeout = history_read_timeout
        self.adapter = adapter
        self.debug = debug

    async def __aenter__(self):
        await self.client.connect()  # type: ignore
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        await self.client.disconnect()

    async def connect(self) -> None:
        await self.client.connect()  # type: ignore

    async def disconnect(self) -> None:
        await self.client.disconnect()  # type: ignore

    @property
    def is_connected(self) -> bool:
        return self.client.is_connected

    async def beep(self) -> None:
        return await self.__get_interface().beep()

    async def status(self) -> RadonEyeStatus:
        return await self.__get_interface().status()

    async def history(self) -> RadonEyeHistory:
        return await self.__get_interface().history()

    async def set_alarm(
        self,
        enabled: bool,  # even when disabled, we still need to provide alarm configuration
        level: float,  # value in bq/m3 or pci/l
        unit: RadonUnit,  # bq/m3 or pci/l
        interval: int,  # in minutes, app supports 10 mins, 1 hour and 6 hours
    ) -> None:
        return await self.__get_interface().set_alarm(enabled, level, unit, interval)

    async def set_unit(self, unit: RadonUnit) -> None:
        return await self.__get_interface().set_unit(unit)

    def __get_interface(self) -> RadonEyeInterface:
        if self.interface:
            return self.interface

        for InterfaceClass in [InterfaceV2, InterfaceV1]:
            interface = InterfaceClass(
                client=self.client,
                status_read_timeout=self.status_read_timeout,
                history_read_timeout=self.history_read_timeout,
                debug=self.debug,
            )
            if interface.supports():
                self.interface = interface
                return interface

        raise NotImplementedError("Not supported device")
